#include "hyp2f1.h"

// winding number = 2
int hasAlmostW2 (cplx z, double delta)
{
  double x;
  double y;
  double xLeft;
  double xRight;
  double yCoord;

  x = real(z);
  y = fabs(imag(z));
  xLeft = 0.82842712474619010; // = (2.0 * sqrt(2.0) - 2.0);
  xRight = 2.0;
  if (x < xLeft - delta || xRight + delta < x)
    return(0);
  yCoord = (double)(sqrt(fabs( -x * x - 4.0 * sqrt(5.0 - 2.0 * x) + 8.0 )));
  if (yCoord + delta <= y)
    return(0);
  return(1);
}

// Gosper's method
// http://www.math.utexas.edu/pipermail/maxima/2006/000126.html
// essential recursion is giveb below
cplx Gosper_2F1 (cplx a, cplx b, cplx c, cplx z)
{
  int k;
	cplx K;
  int kMax;
  cplx d_k;
  cplx e_k;
  cplx f_k;
  cplx d_k1;
  cplx e_k1;
  cplx f_k1, t;
  cplx xi;
  cplx lambda;
  cplx mu;
  cplx result;

  if (ABS(z) <= RADIUS)
  {
    result = series_2F1(a, b, c, z);
    return(result);
  }

  kMax = 2000; // never reached
  xi = z / (z - ONE);
  mu = c - a - b;
  d_k = ZERO;
  e_k = ONE;
  f_k = ZERO;
  for (k = 0; k <= kMax; k++)
  {
		K = cplx(double(k), 0.0);
    lambda = (K + a) * (K + b) / (K + ONE) / (2.0 * K + c) / (2.0 * K + c + ONE);
    d_k1 = (lambda * z * (xi * d_k * (mu + K) + e_k));
    e_k1 = (lambda * z * (-a * b * xi * d_k + (K + c) * e_k));
		t    =  d_k * (-K / (-ONE + z) - (b * a - (mu + K) * K) / (2.0 * K + c) * xi) + e_k;
    f_k1 = (f_k + t);
    //if (areEqual(f_k1, f_k) == 1 && 3 < k)
    if (f_k1 == f_k && 1 < k)
      break;
    d_k = (d_k1);
    e_k = (e_k1);
    f_k = (f_k1);
  }
  result = f_k;

	if ( (ABS(z) < 1.0) &&
			(imag(a) == 0.0) && (imag(b) == 0.0) && (imag(c) == 0.0) && (imag(z) == 0.0) )
		result = real(result) + 0.0*theI;
  
	return(result);
}

/*
                      [d , e , f ] = [0, 1, 0],
                        0   0   0

              [                              (k + c - b - a) d  z        ]
              [                                               k          ]
              [      (k + a) (k + b) z (e  - --------------------)       ]
              [                          k          1 - z                ]
              [      ---------------------------------------------       ]
              [                             c       c + 1                ]
              [              4 (k + 1) (k + -) (k + -----)               ]
              [                             2         2                  ]
 [ d      ]   [                                                          ]
 [  k + 1 ]   [                           a b d  z                       ]
 [        ]   [                                k                         ]
 [ e      ] = [        (k + a) (k + b) z (-------- + (k + c) e )         ] 
 [  k + 1 ]   [                            1 - z              k          ]
 [        ]   [        -----------------------------------------         ]
 [ f      ]   [                             c       c + 1                ]
 [  k + 1 ]   [              4 (k + 1) (k + -) (k + -----)               ]
              [                             2         2                  ]
              [                                                          ]
              [      d  (k ((c - b - a) z + k (z - 2) - c) - a b z)      ]
              [       k                                                  ]
              [ f  - ---------------------------------------------- + e  ]
              [  k                        c                            k ]
              [                    2 (k + -) (1 - z)                     ]
              [                           2                              ]

Then d_k and e_k approach 0 like (z^2/(4z-4))^k, and the running sum f_k
approaches 2F1(a,b;c|z)
*/