#ifndef _HYP2F1_H_
#define _HYP2F1_H_


#include <cmath>
#include <complex>
using namespace std;

#define DBL_EPSILON       2.2204460492503131e-16    // float.h, 2^(-52)
#define DBL_MIN           2.2250738585072014e-308   // float.h, 2^(-1021)
#define SQRT_DBL_EPSILON  1.4901161193847656e-08    // 2^(-26)
#define PI                3.14159265358979324
#define sqrt2Pi           2.50662827463100050       // = sqrt(2*Pi)

#define switchGAMMA       20  // switching between recursion and reflection

#define RADIUS            0.9 // 1843.0/2048.0 // ~ 0.8999... for hypgeom series
#define STEPSIZE          3.0/1024.0    // ~ 0.00268... numerical differentiation

#define regRADIUS         1.0/1024.0    // for averaging around (numerical) poles

typedef std::complex<double> cplx;

#define ZERO              cplx(0.0, 0.0)
#define ONE               cplx(1.0, 0.0)
#define theI              cplx(0.0, 1.0)

// the 8 unit roots on the unitCircle * 4*regRADIUS, ignoring symmetries ...
static cplx ROOTS[] = {
  cplx(+1.0                  * 4.0*regRADIUS, +0.0                 * 4.0*regRADIUS),
  cplx(+0.707106781186547525 * 4.0*regRADIUS, +.707106781186547525 * 4.0*regRADIUS),
  cplx(+0.0                  * 4.0*regRADIUS, +1.0                 * 4.0*regRADIUS),
  cplx(-0.707106781186547525 * 4.0*regRADIUS, +.707106781186547525 * 4.0*regRADIUS),
  cplx(-1.0                  * 4.0*regRADIUS, +0.0                 * 4.0*regRADIUS),
  cplx(-0.707106781186547525 * 4.0*regRADIUS, -.707106781186547525 * 4.0*regRADIUS),
  cplx(+0.0                  * 4.0*regRADIUS, -1.0                 * 4.0*regRADIUS),
  cplx(+0.707106781186547525 * 4.0*regRADIUS, -.707106781186547525 * 4.0*regRADIUS)
  };
#define degROOTS  8 // = number of entries, should be chosen even

#define RE(z)     ((double)(std::real((cplx)z)))
#define IM(z)     ((double)(std::imag((cplx)z)))
#define ABS(z)    ((double)(std::abs((cplx)z)))
#define ARG(z)    ((double)(std::arg((cplx)z)))
//#define CONJUGATE(z) ((cplx)(std::conj((cplx)z)))

#define SIN(z)    ((cplx)(std::sin((cplx)z)))
#define EXP(z)    ((cplx)(std::exp((cplx)z)))
#define LOG(z)    ((cplx)(std::log((cplx)z)))
#define POW(z,a)  ((cplx)(std::pow((cplx)z,(cplx)a)))


// numTools
double maxNORM (cplx z);

int areClose (const double x1, const double x2, const double epsilon);
int areEqual (cplx a, cplx b);

int roundToInteger (double x);
int isNumericalInteger (cplx z);
int isNumericalNaturalNumber (cplx z);
int isOdd (int m);

//int isAlmostReal (cplx z);
//int isAlmostInteger (cplx z, double epsilon);
//int isAlmostNaturalNumber (cplx z, double epsilon);

double moduloPI(double x);
cplx rSIN (cplx mu);


// gamma
cplx pochhammer(cplx z, int n);
cplx GAMMA(cplx z);
cplx rGAMMA (cplx a);

// seriesCase
cplx series_2F1 (cplx A, cplx B, cplx C, cplx Z);
cplx polynomialseries_2F1 (cplx a, cplx b, cplx c, cplx z);
cplx polynom_2F1 (cplx a, cplx b, cplx c, cplx z, 
  int deg);

// gosper
int hasAlmostW2 (cplx z, double delta);
cplx Gosper_2F1 (cplx a, cplx b, cplx c, cplx z);

// linTransforms
cplx Euler (cplx a, cplx b, cplx c, cplx z);
cplx Euler_polynom (cplx a, cplx b, cplx c, cplx z);
cplx Pfaff (cplx a, cplx b, cplx c, cplx z);
cplx AS_15_3_6_Leb (cplx a, cplx b, cplx c, cplx z);
cplx AS_15_3_7_Leb (cplx a, cplx b, cplx c, cplx z);
cplx AS_15_3_9_Leb (cplx a, cplx b, cplx c, cplx z);

// exceptionals
cplx exceptional_15_3_6 (cplx a, cplx b, cplx c, cplx z);
cplx exceptional_15_3_7 (cplx a, cplx b, cplx c, cplx z);
cplx exceptional_15_3_9 (cplx a, cplx b, cplx c, cplx z);

// the actual hypergeometric functions 2F1
cplx G_2F1_regularized (cplx a, cplx b, cplx c, cplx z);
cplx G_2F1 (cplx a, cplx b, cplx c, cplx z);

cplx hyp2F1 (cplx a, cplx b, cplx c, cplx z);
cplx hyp2F1_regularized (cplx a, cplx b, cplx c, cplx z);

// interface, structure for Excel
typedef struct
{
  double re,im;
}
cplx_structure;

#endif // _HYP2F1_H_