#include "hyp2f1.h"

cplx branchPoint_2F1 (cplx a, cplx b, cplx c)
{
  // branching point z = 1:
  //if (0.0 == IM(z) && 1.0 == RE(z))
  {
    //care only for the classical case
    if ( 0.0 < RE(c - a - b) )
      return(GAMMA(c) * GAMMA(c - a - b) * rGAMMA(c - a) * rGAMMA(c - b));
    // otherwise imitate limit from the left by smallest IEEE number < 1
    hyp2F1(a,b,c, cplx(1.0 - DBL_EPSILON*0.5, 0.0));
  }
  return ZERO;
}

cplx G_2F1 (cplx a, cplx b, cplx c, cplx z)
{
  cplx result;
  double delta;
  cplx w;
  cplx M;

  delta = 0.51776695296636881e-1; // = (sqrt(2.0) - 1.0) / 8.0

  if (areEqual(z, ZERO) == 1)
    return(ONE);

  if (isNumericalNaturalNumber(-c) == 1)
    return(0.1e308);

  if ( ABS(b) < ABS(a) )
    return G_2F1(b,a,c,z);

  // try to find good constellations for parameters by Euler's transform
  // note: applying twice gives identity (---> infinite loop by recursion),
  // so care for strict conditions to apply only once
  if ( ABS((c-a)*(c-b)/c) < ABS(a*b/c) )
    return Euler(a,b,c, z); // w = z, it is not changed

  // Using Pfaff's transform is another way, but I hesitate:
  // it changes w = z/(z-1) and parameter constellation should be controlled ...
  // may be demanding abs( z/(z-1) ) <= abs(z) ? 
  // ---> ]-oo, ..., 0[ und [2, ... , oo[ falls reel
  
  if (isNumericalNaturalNumber(-a) == 1 || isNumericalNaturalNumber(-b) == 1)
    return(polynomialseries_2F1(a, b, c, z));
  
  if (isNumericalNaturalNumber(-(c - a)) == 1 || isNumericalNaturalNumber(-(c - b)) == 1)
    return(Euler_polynom(a, b, c, z)); // Euler's relation gives a polynom for that

  // reduction to some other elementary function, deactivate it for tests,
  // since test function often rely on that relations
  // power: (1-x)^(-a) = hypergeom([a],[],x)
  // log: ln(1-x) = -x*hypergeom([1, 1],[2],x)
  // asin: arcsin(x^(1/2)) = x^(1/2)*hypergeom([1/2, 1/2],[3/2],x)
  // atan: arctan(x^(1/2)) = x^(1/2)*hypergeom([1/2, 1],[3/2],-x)
  /*
  if (b == c) return POW(ONE - z, -a);
  if (a == c) return POW(ONE - z, -b);
  if (ONE == a && ONE == b && c == ONE+ONE) return -log(ONE - z)/z;
  */

  // branching point z = 1:
  if (0.0 == IM(z) && 1.0 == RE(z))
  {
    //care only for the classical case
    if ( 0.0 < RE(c - a - b) )
      return(GAMMA(c) * GAMMA(c - a - b) * rGAMMA(c - a) * rGAMMA(c - b));
    // otherwise imitate limit from the left by smallest IEEE number < 1
    G_2F1(a,b,c, cplx(1.0 - DBL_EPSILON*0.5, 0.0));
  }

  // branch cut: from below
  if ( 0.0 == IM(z) && 1.0 < RE(z) )
    z = RE(z) - 1e-306 * theI;


  if ( ABS(z) <= RADIUS )
    return series_2F1(a,b,c,z);

  if ( 2.0 <= ABS(z) )
    return AS_15_3_7_Leb(a,b,c,z);      // w = 1 / z;

  if ( ABS(z - ONE) <= 1.0/1024.0 )
    return(AS_15_3_6_Leb(a, b, c, z));  // w = 1 - z;

  // = AS_15_3_8 = AS_15_3_7 @ Pfaff, w = 1 / [z/(z-1)] = 1/(z-1)
  if ( 0.0 <= RE(z) && 
    RADIUS < ABS(z) && ABS(z) <= 3.0 &&
    fabs(ARG(z)) <= 1.0/3.0 ) // && 1.0/RADIUS < ABS(z/(z-ONE)) )
    if ( ABS(a*(c - b)) < ABS((c - a)*b) )            // use A&S 15.3.4
      return POW(ONE - z, - a) * AS_15_3_7_Leb(a, c - b, c, z/(z-ONE));
    else                                              // use A&S 15.3.5
      return POW(ONE - z, - b) * AS_15_3_7_Leb(c - a, b, c, z/(z-ONE));

  return Gosper_2F1(a, b, c, z);


  // use Pfaff's transform with w = z/(z-1) for the negative half plane,
  // which does not have parameter restrictions
  // note: for the choosen restriction we do not fall into W2 or 
  // arive within the radius for series computation, we will only miss
  // those z=r*exp(I*Pi*phi) with R/(R-1) = 9.0 < r, 0 <= phi <= 1
  if ( ( RE(z) <= 0.0 && 1.0 <= ABS(z) && ABS(z) <= 3.0 )
      || (ABS(z/(z-ONE)) < RADIUS ) 
      )
    return Pfaff(a,b,c, z);             // w = z/(z-1)

  if ( 1.0/RADIUS <= ABS(z) )
    return AS_15_3_7_Leb(a,b,c,z);      // w = 1 / z;

  if ( 0.0 <= RE(z) && fabs(ARG(z)) <= 1.0/3.0 && 1.0/RADIUS < ABS(z/(z-ONE)) ) 
    return Pfaff(a,b,c, z);             // w = z/(z-1)

  // Vorsicht mit Pfaff: Pfaff^2 = identity
 
  //if ( ABS(z) < 0.9 && ABS(z - ONE) < RADIUS )        // AVt, 02.11.09
    //return(AS_15_3_6_Leb(a, b, c, z));  // w = 1 - z;

  //if ( ABS(z) < 0.99 && ABS((z - ONE)/z) < RADIUS )
    //return(AS_15_3_9_Leb(a, b, c, z));  // w = 1 - 1 / z; ???
 

  

  //if (hasAlmostW2(z, delta) == 1 && 0.5 < RE(z) && ABS(z) <= 1.5)
    //return (AS_15_3_6_Leb(a, b, c, z));  // w = 1 - z;

  if (hasAlmostW2(z, delta) == 1)
    return(AS_15_3_9_Leb(a, b, c, z));  // w = 1 - 1 / z;
  
  result = Gosper_2F1(a, b, c, z);
  return(result);
}
