#include "hyp2f1.h"
#pragma fp_contract (off)

cplx polynom_2F1 (cplx a, cplx b, cplx c, cplx z,
  int deg)
{
  int n;
  int nMax;
  cplx r;
  cplx s;
  cplx t;

  if (2000 < deg)
    nMax = 2000;
  else
    nMax = deg;

  if (nMax == 0)
    return(ONE);
  if (areEqual(a, ZERO) == 1 || areEqual(b, ZERO) == 1)
    return(ONE);
  if (areEqual(z, ZERO) == 1)
    return(ONE);
  
  if (nMax == 1)
    return(ONE + a * b / c * z);

  n = 0;
  s = ONE;
  t = ONE;
  while (n < nMax)
  {
    n = n + 1;
    r = a * b / c / (cplx) n * z;
    if (areEqual(r, ZERO) == 1)
      break;

    t = t * r;
    s = t + s;
    a = a + ONE;
    b = b + ONE;
    c = c + ONE;
  }

  // clean up for the real case
  if ( (IM(a) == 0.0) && (IM(b) == 0.0) && (IM(c) == 0.0) &&
        (IM(z) == 0.0) )
    s = RE(s) + 0.0*theI;

  return(s);
}

// to be used, if a or b are negative integers, then it is a polynom
// and one has to determine the degree
cplx polynomialseries_2F1 (cplx a, cplx b, cplx c, cplx z)
{
  int n=0;
  int int_a=0, int_b=0;
  cplx result=ZERO;
  double re_a = RE(a), re_b = RE(b);
  
  int_a = roundToInteger( re_a );
  int_b = roundToInteger( re_b );
  
  // a integer and <= 0
  if ( int_a <= 0 && IM(a) == 0.0 && re_a == (double)int_a )
    n = int_a;    // i.e. a = n
  // b integer and <= 0
  if ( int_b <= 0 && IM(b) == 0.0 && re_b == (double)int_b
    && int_b <= n ) 
    n = int_b;    // b = n
  // the smaller one is taken, n <= 0
  
  result = polynom_2F1(a, b, c, z, -n);
  return(result);
}

// hypergeometric series for abs(z) < 1
cplx series_2F1 (cplx a, cplx b, cplx c, cplx z)
{
  int n;
  int nMax;
  cplx r;
  cplx s;
  cplx sNew;
  cplx t;
  int lastEqual = 0;

  nMax = 2000; // this is never reached

  if ( a == ZERO || b == ZERO )
    return(ONE);
  if (z == ZERO)
    return(ONE);

  n = 0;
  s = ONE;
  t = ONE;
  while (n < nMax)
  {
    n = n + 1;
    r = ( a * (b/c) ) / (cplx) n * z;
    t = (t * r);
    sNew = (t + s);

    // check 2 subsequent terms being equal
    if( s == sNew ) 
      if ( 1 == lastEqual ) break;
      else lastEqual = 1;
    else
      lastEqual = 0; 
    s = sNew;
    
    a = (a + ONE);
    b = (b + ONE);
    c = (c + ONE);
  }

  // clean up
  if ( (IM(a) == 0.0) && (IM(b) == 0.0) && (IM(c) == 0.0) &&
        (IM(z) == 0.0) )
    s = RE(s) + 0.0*theI;

  return(s);
}
